import java.awt.*;
import java.awt.image.*;
import java.awt.event.*;
import java.util.*;
import Mandelbrot;
import Julia;
import Rect;
import RoundedNumber;
import Selection;
import SizedStack;

public class FractalCanvas extends java.awt.Canvas
implements MouseListener, MouseMotionListener, ComponentListener {

    protected static final int draggingEdge = 1, draggingCorner = 5,
                             draggingAll = 9;
    
    protected boolean enabled;
    protected boolean initialized;
    protected int margin;
    protected double aspectRatio;
    protected MemoryImageSource buffimgsrc;
    protected Image buffimage;
    protected Fractal fractal;
    protected java.awt.Label[] boundLabels;
    protected FractalApplet thisApplet;
    protected int draggingWhat;
    protected int dragStartX, dragStartY;
    protected Selection originalSelection;
    protected Selection selection;
    protected Rect logicalBounds;
    protected Rect screenBounds;
    protected boolean quickRender;
    protected SizedStack zoomHistory;
    
    public FractalCanvas() {
        super();
        initialized = false;
        buffimgsrc = null;
        buffimage = null;
        selection = new Selection();
        logicalBounds = new Rect();
        screenBounds = new Rect();
        margin = 5;
        quickRender = true;
        zoomHistory = new SizedStack(5);
	enabled = true;
       
        addMouseListener (this);
        addMouseMotionListener (this);
        addComponentListener (this);
     }
    
    public FractalCanvas(Fractal frac, FractalApplet _applet, java.awt.Label xmax,
    java.awt.Label xmin, java.awt.Label ymax, java.awt.Label ymin) {
        this();
        fractal = frac;
        thisApplet = _applet;
        boundLabels = new java.awt.Label[4];
        boundLabels[0] = xmin;
        boundLabels[1] = xmax;
        boundLabels[2] = ymin;
        boundLabels[3] = ymax;
    }
    
    public void setEnabled(boolean en) {
       enabled = en;
    }
   
    private void initialize() { 
        if (fractal == null)
            fractal = new Mandelbrot();
        fractal.setBounds(fractal.getAllBounds());
        componentResized((ComponentEvent)null);
        initialized = true;
    }
    
    public void componentResized(ComponentEvent evt) {
	if (!enabled)
	 	return;
        screenBounds.setXmin(margin);
        screenBounds.setYmin(margin);
        screenBounds.setXmax(getSize().width - screenBounds.getXmin() - margin);
        screenBounds.setYmax(getSize().height - screenBounds.getYmin() - margin);
        aspectRatio = screenBounds.getWidth() / screenBounds.getHeight();
        selection.setScreenBounds(screenBounds);
        selection.setAspectRatio(aspectRatio);
        if (fractal != null)
            setFractalBounds(getFractalBounds());
    }

    public void update(Graphics g) {
       if (enabled) 
	       paint(g);
    }
    
    public void paint(Graphics g) {
        if (!initialized)
            initialize();
        if (fractal.needsRegeneration()) {
            thisApplet.setEnabled(false);
            grayOutImage(g);
            buffimgsrc = new MemoryImageSource((int)screenBounds.getWidth(),
                                               (int)screenBounds.getHeight(),
                                               fractal.render((int)screenBounds.getWidth(),
                                                                (int)screenBounds.getHeight(),
                                                                quickRender),
                                               0, (int)screenBounds.getWidth());
            buffimage = createImage(buffimgsrc);
            for (int i=0; i<4; i++)
                boundLabels[i].setText(new RoundedNumber(logicalBounds.getRect()[i]).toString());
            thisApplet.setEnabled(true);
        }
        g.drawImage(buffimage, (int)screenBounds.getXmin(), (int)screenBounds.getYmin(), this);
        selection.repaint(g);
    }

    public void grayOutImage(Graphics g) {
        g.setColor(new Color(0xcdcdee));
        for (int y=(int)screenBounds.getYmin(); y < screenBounds.getYmax(); y++) {
            if (y % 3 == 0)
                g.drawLine((int)screenBounds.getXmin(), y, (int)screenBounds.getXmax()-1, y);
        }
    }

    public void setAxisLabels(java.awt.Label x, java.awt.Label y) {
        fractal.setAxisLabels(x, y);
    }
    
    public int getIterations() {
        if (fractal != null) {
            if (fractal instanceof Mandelbrot)
                return (((Mandelbrot)fractal).getIterations());
            else if (fractal instanceof Julia)
                return (((Julia)fractal).getIterations());
        }
        return (0);
    }
        
    
    public void setIterations(int i) {
        if (!enabled)
	 	return;
       	if (fractal != null) {
            if (fractal instanceof Mandelbrot)
                ((Mandelbrot)fractal).setIterations(i);
            else if (fractal instanceof Julia)
                ((Julia)fractal).setIterations(i);
        }
    }
    
    public void setFractalBounds(Rect bounds) {
	if (!enabled)
	 	return;
       if (!zoomHistory.empty()) {
            if (!((Rect)zoomHistory.peek()).equals(fractal.getBounds()))
                zoomHistory.push(fractal.getBounds());
        } else {
            zoomHistory.push(fractal.getBounds());
        }
        setFractalBoundsSkipHistory(bounds);
    }
    
    public void setFractalBoundsSkipHistory(Rect bounds) {
	if (!enabled)
	 	return;
        bounds.applyAspectRatio(aspectRatio);
        fractal.setBounds(bounds);
        logicalBounds = fractal.getBounds();
        selection.setLogicalBounds(logicalBounds);
        selection.setSelection(logicalBounds);
	thisApplet.setStatusBar(fractal.toString() + " (" + Math.round(fractal.getAllBounds().getWidth() / bounds.getWidth()) + "x)");
        repaint();
    }
    
    public Fractal getRenderer() {
        return (fractal);
    }
    
    public Rect getFractalBounds() {
        return (fractal.getBounds());
    }
    
    public Rect getNewBounds() {
        return (new Rect(selection.getSelection()));
    }
    
    public void zoomAll() {
	if (!enabled)
	 	return;
       	Rect bounds = fractal.getAllBounds();
        setFractalBounds(bounds);
    }
    
    public void zoomLast() {
	if (!enabled)
	 	return;
       if (!zoomHistory.empty()) {
            setFractalBoundsSkipHistory((Rect) zoomHistory.pop());
        }
    }
    
    public void clearZoomHistory() {
        zoomHistory.clear();
    }
    
    public Rect getPoint() {
        Rect sel = selection.getSelection();
        double x = sel.getX() + sel.getWidth()/2;
        double y = sel.getY() + sel.getHeight()/2;
        Rect point = new Rect(x, x, y, y);
        return (point);
    }
    
    public void mouseMoved(MouseEvent evt) {
	if (!enabled)
	 	return;
       	int x = evt.getX();
        int y = evt.getY();
        
        Rect sel = selection.getTranslatedSelection();
        
        if (sel.contains(x, y)) {
            setCursor(new Cursor(Cursor.MOVE_CURSOR));
        } else if (Math.abs(sel.getXmin() - x) < 10 && Math.abs(sel.getYmin() - y) < 10) {
            setCursor(new Cursor(Cursor.SW_RESIZE_CURSOR));
        } else if (Math.abs(sel.getXmax() - x) < 10 && Math.abs(sel.getYmin() - y) < 10) {
            setCursor(new Cursor(Cursor.SE_RESIZE_CURSOR));
        } else if (Math.abs(sel.getXmin() - x) < 10 && Math.abs(sel.getYmax() - y) < 10) {
            setCursor(new Cursor(Cursor.NW_RESIZE_CURSOR));
        } else if (Math.abs(sel.getXmax() - x) < 10 && Math.abs(sel.getYmax() - y) < 10) {
            setCursor(new Cursor(Cursor.NE_RESIZE_CURSOR));
        } else if (Math.abs(sel.getXmin() - x) < 10) {
            setCursor(new Cursor(Cursor.W_RESIZE_CURSOR));
        } else if (Math.abs(sel.getXmax() - x) < 10) {
            setCursor(new Cursor(Cursor.E_RESIZE_CURSOR));
        } else if (Math.abs(sel.getYmin() - y) < 10) {
            setCursor(new Cursor(Cursor.S_RESIZE_CURSOR));
        } else if (Math.abs(sel.getYmax() - y) < 10) {
            setCursor(new Cursor(Cursor.N_RESIZE_CURSOR));
        } else {
            setCursor(new Cursor(Cursor.DEFAULT_CURSOR));
            /* In JDK 1.2 SetCursor(null) can be used. JDK 1.1, however, does not support that */
        }
    }
    
    public void mousePressed(MouseEvent evt) {
        if (!enabled)
	 	return;
       	int x = evt.getX();
        int y = evt.getY();
        
        Rect sel = selection.getTranslatedSelection();
        
        if (sel.contains(x, y)) {
            draggingWhat = draggingAll;
        } else if (Math.abs(sel.getXmin() - x) < 10 && Math.abs(sel.getYmin() - y) < 10) {
            draggingWhat = draggingCorner + 0;
        } else if (Math.abs(sel.getXmax() - x) < 10 && Math.abs(sel.getYmin() - y) < 10) {
            draggingWhat = draggingCorner + 1;
        } else if (Math.abs(sel.getXmin() - x) < 10 && Math.abs(sel.getYmax() - y) < 10) {
            draggingWhat = draggingCorner + 2;
        } else if (Math.abs(sel.getXmax() - x) < 10 && Math.abs(sel.getYmax() - y) < 10) {
            draggingWhat = draggingCorner + 3;
        } else if (Math.abs(sel.getXmin() - x) < 10) {
            draggingWhat = draggingEdge + 0;
        } else if (Math.abs(sel.getXmax() - x) < 10) {
            draggingWhat = draggingEdge + 1;
        } else if (Math.abs(sel.getYmin() - y) < 10) {
            draggingWhat = draggingEdge + 2;
        } else if (Math.abs(sel.getYmax() - y) < 10) {
            draggingWhat = draggingEdge + 3;
        } else {
            draggingWhat = 0;
        }
        
        originalSelection = new Selection(selection);
        dragStartX = evt.getX();
        dragStartY = evt.getY();
    }
    
    public void mouseDragged(MouseEvent evt) {
	if (!enabled)
	 	return;
       	int x = evt.getX();
        int y = evt.getY();
        int distanceX = x - dragStartX;
        int distanceY = y - dragStartY;
        
        Selection proposed = new Selection(originalSelection);
        
        if (draggingWhat == draggingCorner+0) {
            proposed.resizeUsingScreenCoordinates(-distanceX + distanceY);
        } else if (draggingWhat == draggingCorner+1) {
            proposed.resizeUsingScreenCoordinates(distanceX + distanceY);
        } else if (draggingWhat == draggingCorner+2) {
            proposed.resizeUsingScreenCoordinates(-distanceX - distanceY);
        } else if (draggingWhat == draggingCorner+3) {
            proposed.resizeUsingScreenCoordinates(distanceX - distanceY);
        } else if (draggingWhat == draggingEdge+0) {
            proposed.resizeUsingScreenCoordinates(-distanceX);
        } else if (draggingWhat == draggingEdge+1) {
            proposed.resizeUsingScreenCoordinates(distanceX);
        } else if (draggingWhat == draggingEdge+2) {
            proposed.resizeUsingScreenCoordinates(distanceY);
        } else if (draggingWhat == draggingEdge+3) {
            proposed.resizeUsingScreenCoordinates(-distanceY);
        } else if (draggingWhat == draggingAll) {
            proposed.moveUsingScreenCoordinates(distanceX, distanceY);
        }
        selection.setSelection(proposed);
        repaint();
    }

    public void mouseReleased(MouseEvent evt) {}
    public void mouseClicked(MouseEvent evt) {}
    public void mouseEntered(MouseEvent evt) {}
    public void mouseExited(MouseEvent evt) {}

    public void componentHidden(ComponentEvent evt) {}
    public void componentMoved(ComponentEvent evt) {}
    public void componentShown(ComponentEvent evt) {
       repaint();
    }
}
