/**
 * This class is used to render a mandelbrot set.
 * It uses a cache to store generated fractals a few generations back. The
 * cache is implemented as a stack with FractalData objects.
 */

import java.awt.Image;
import java.awt.image.*;
import Rect;
import java.awt.Component;
import Fractal;
import SizedStack;
import FractalData;

public class Mandelbrot extends Fractal{

    protected int maxiter;      //Maximum iterations for the algorithm
    protected boolean touched;  //Needs to be rerendered
    protected FractalData fractalData;  //Buffer to hold the rendered fractal
    protected SizedStack cache;         //Cache for generated fractals

    public Mandelbrot() {
	super();
        all_bounds = new Rect(-2.1, 0.7, -1.2, 1.2);
        bounds = new Rect(all_bounds);
        maxiter = 50;
        touched = true;
	cache = new SizedStack(5);
	fractalData = new FractalData();
    }
    
    public boolean needsRegeneration() {
        if (touched)
            return true;
        else
            return false;
    }

    /**
     * Generates a mandelbrot fractal. Calls renderEscapeIter().
     * @param width int width
     * @param height int height
     * @param quick not used
     * @return int[] a pixel-array
     */
   public int[] render(int width, int height, boolean quick) {
	return(renderEscapeIter(width, height, quick));
    }
   
    /**
     * Generates a mandelbrot fractal.
     */
    public int[] renderEscapeIter(int width, int height, boolean quick) {

        fractalData = new FractalData(width, height, maxiter, bounds);
       
	/* Have a look in the cache */
        if (cache.contains(fractalData)) {
	    touched = false;
	    return (((FractalData)cache.getElement(cache.indexOf(fractalData))).buff);
	}

        int[] color = new int[3];
        int i;
        double a, b, c, d, e;
        
        for (int yp=0; yp < height; yp++) {
            for (int xp=0; xp < width; xp++) {
                c = ((double)xp/width) * (bounds.get(Rect.XMAX)-bounds.get(Rect.XMIN)) + bounds.get(Rect.XMIN);
                d = (1-(double)yp/height) * (bounds.get(Rect.YMAX)-bounds.get(Rect.YMIN)) + bounds.get(Rect.YMIN);
                          
                a = b = 0;
                for (i=1; i < maxiter; i++) {
                    e = a;
                    a = a*a - b*b;
                    b = 2*e*b;
                    a+=c;
                    b+=d;
                    if (Math.sqrt(a*a + b*b) > 2)
                        break;
                }
                
                if (i >= maxiter) {
                    color[0] = color[1] = color[2] = 0;
                } else {
		    //A colour scheme of black->red->orange->yellow->white is used
		    color[0] = (int)((double)i/maxiter * 255 * 3);
                    color[1] = (int)((double)i/maxiter * 255 * 3/2);
                    color[2] = (int)((double)i/maxiter * 255 * 1);
                    if (color[0] > 255) color[0] = 255;
                    if (color[1] > 255) color[1] = 255;
                    if (color[2] > 255) color[2] = 255; //shouldn't happen
                }
                
                fractalData.buff[(yp * width) + xp] = (255 << 24) | (color[0] << 16) |
                    (color[1] << 8) | (color[2]);
            }
        }
        touched = false;
	cache.push(fractalData);
        return (fractalData.buff);
    }

    /**
     * Generates a mandelbrot fractal with the Distance Estimator method.
     * Inspired by http://www.mrob.com
     */
    public int[] renderDistanceEstimator(int width, int height, boolean quick) {
        fractalData = new FractalData(width, height, maxiter, bounds);
        
	/* Have a look in the cache */
        if (cache.contains(fractalData)) {
	    touched = false;
	    return (((FractalData)cache.getElement(cache.indexOf(fractalData))).buff);
	}

        int[] color = new int[3];
        int i;
        double cre, cim, zre, zim, dzre, dzim, z2re, z2im, tmp;
        double zabs, dzabs;
        int res;
       
        for (int yp=0; yp < height; yp++) {
            for (int xp=0; xp < width; xp++) {
                cre = ((double)xp/width) * (bounds.get(Rect.XMAX)-bounds.get(Rect.XMIN)) + bounds.get(Rect.XMIN);
                cim = (1-(double)yp/height) * (bounds.get(Rect.YMAX)-bounds.get(Rect.YMIN)) + bounds.get(Rect.YMIN);
                
                zre = zim = 0;
	        dzre = dzim = 0;
                for (i=1; i < maxiter; i++) {
                    //z2 = z*z + c
                    z2re = zre*zre - zim*zim + cre;
                    z2im = 2*zre*zim + cim;
		    //dz = 2 * z * dz + 1
		    tmp = dzre;
		    dzre = 2*zre*dzre - 2*zim*dzim + 1;
		    dzim = zre*dzim + zim*tmp;
		    //z = z2
		    zre = z2re; zim = z2im;
                    if (Math.sqrt(zre*zre + zim*zim) > 2)
                        break;
                }
                
	        zabs = Math.sqrt(zre*zre + zim*zim);
	        dzabs = Math.sqrt(dzre*dzre + dzim*dzim);
	        res = (int)(Math.log(zabs*zabs) * zabs/dzabs);

                if (i >= maxiter) {
                    color[0] = color[1] = color[2] = 0;
                } else {
		    color[0] = (int)((double)i/maxiter * 255 * 2);
                    color[1] = (int)((double)i/maxiter * 255 * 1);
                    color[2] = (int)((((double)res+20)/70)*255);
                    if (color[0] > 255) color[0] = 255;
                    if (color[1] > 255) color[1] = 255;
                    if (color[2] > 255) color[2] = 255; //shouldn't happen
                }
                
                fractalData.buff[(yp * width) + xp] = (255 << 24) | (color[0] << 16) |
                    (color[1] << 8) | (color[2]);
            }
        }
        touched = false;
	cache.push(fractalData);
        return (fractalData.buff);
    }

    public void setBounds(Rect b) {
        touched = true;
        bounds = new Rect(b);
    }

    public int getIterations() {
        return (maxiter);
    }
    
    public void setIterations(int i) {
        maxiter = i;
    }
    
    public Rect getBounds() {
        return new Rect(bounds);
    }
    
    public Rect getAllBounds() {
        return (new Rect(all_bounds));
    }
    
    //Used for debuging
    public String toString() {
        return ("Mandelbrot fractal");
    }
    
    public void setAxisLabels(java.awt.Label x, java.awt.Label y) {
        x.setText("Re c");
        y.setText("Im c");
    }
    
}
