/* Represents a rectangle. Maximum and minimum X and Y coordinates are stored
 * and width and height may be calculated. An aspect ratio may be applied to the
 * rectangle.
 */

public class Rect {

    protected double[] rect;
    public final static int XMIN = 0;
    public final static int XMAX = 1;
    public final static int YMIN = 2;
    public final static int YMAX = 3;

    public Rect() {
        rect = new double[4];
    }
    
    public Rect(Rect x) {
        this();
	set(x);
    }
    
    public Rect(double xmin, double xmax, double ymin, double ymax) {
        this();
	set(xmin, xmax, ymin, ymax);
    }
    
    public void add(Rect add) {
        rect[XMIN] += add.rect[XMIN];
        rect[XMAX] += add.rect[XMAX];
        rect[YMIN] += add.rect[YMIN];
        rect[YMAX] += add.rect[YMAX];
    }
    
    /* Applies an aspect ratio to the rectangle. The dimensions are calculated
     * to adhere to this ratio: width = height * ratio
     * The width and height are changed in such a manner that none of them
     * grows larger.
     */
    public void applyAspectRatio(double aspectRatio) {
        if ((getWidth() / getHeight()) < aspectRatio) {
            double xdiff = getHeight()*aspectRatio - getWidth();
            rect[XMIN] -= xdiff/2;
            rect[XMAX] += xdiff/2;
        } else {
            double ydiff = getWidth()/aspectRatio - getHeight();
            rect[YMIN] -= ydiff/2;
            rect[YMAX] += ydiff/2;
        }
    }
    
    public void setXmin(double x) {
        rect[XMIN] = x;
    }
    
    public void setXmax(double x) {
        rect[XMAX] = x;
    }
    
    public void setYmin(double x) {
        rect[YMIN] = x;
    }
    
    public void setYmax(double x) {
        rect[YMAX] = x;
    }
    
    public void setPos(double x, double y) {
        rect[XMAX] += x - rect[XMIN];
        rect[XMIN] = x;
        rect[YMAX] += y - rect[YMIN];
        rect[YMIN] = y;
    }
    
    public void set(Rect x) {
        double[] y = x.getRect();
	if (x != null)
	    for (int i=0; i < 4; i++)
                rect[i] = y[i];
    }
    
    public void set(double xmin, double xmax, double ymin, double ymax) {
        rect[XMIN] = xmin;
        rect[XMAX] = xmax;
        rect[YMIN] = ymin;
        rect[YMAX] = ymax;
    }
    
    public void set(double xmin, double xmax, double ymin, double ymax,
    double aspectRatio) {
        set(xmin, xmax, ymin, ymax);
        applyAspectRatio(aspectRatio);
    }
    
    public double[] getRect() {
        return (rect);
    }

    public double getX() {
        return (rect[XMIN]);
    }
    
    public double getY() {
        return (rect[YMIN]);
    }
    
    public double getXmin() {
        return (rect[XMIN]);
    }
    
    public double getXmax() {
        return (rect[XMAX]);
    }
    
    public double getYmin() {
        return (rect[YMIN]);
    }
    
    public double getYmax() {
        return (rect[YMAX]);
    }
    
    public double getWidth() {
        return (rect[XMAX] - rect[XMIN]);
    }
    
    public double getHeight() {
        return (rect[YMAX] - rect[YMIN]);
    }
    
    public double get(int i) {
        return (rect[i]);
    }

    /* returns true if the given coordinate is inside the rectangle */
    public boolean contains(int x, int y) {
        double y1 = rect[YMIN], y2 = rect[YMAX], x1 = rect[XMIN], x2 = rect[XMAX];
        if (y1 > y2) {
            double tmp = y2;
            y2 = y1;
            y1 = tmp;
        }
        if (x1 > x2) {
            double tmp = x2;
            x2 = x1;
            x1 = tmp;
        }
        return (x >= x1 && x <= x2 && y >= y1 && y <= y2);
    }
    
    public boolean equals(Rect other) {
        return ((rect[XMIN] == other.rect[XMIN]) && (rect[XMAX] == other.rect[XMAX]) &&
                (rect[YMIN] == other.rect[YMIN]) && (rect[YMAX] == other.rect[YMAX]));
    }
   
    public String toString() {
        return ("xmin: " + rect[XMIN] + " xmax: " + rect[XMAX] + 
                " ymin: " + rect[YMIN] + " ymax: " + rect[YMAX]);
    }
}
