/* Represents a selected area of a fractal. The difficulties are that
 * coordinates can be given as screen coordinates (ie. in pixels, the upper
 * left corner being (0,0)) or as logical coordinates (ie. the input to the
 * algorithm calculating the fractal). This class is able to translate
 * between the two formats and move or resize a selection.
 * May contain a few bugs
 */

import java.awt.*;
import Rect;

public class Selection {

    protected Rect selection;
    protected Rect logicalBounds;
    protected Rect screenBounds;
    protected Rect translatedSelection;
    protected double aspectRatio;
    
    public Selection() {
        selection = new Rect();            //The selected area, logical coordinates
        logicalBounds = new Rect();        //Image bounds, logical coordinates
        screenBounds = new Rect();         //Image bounds, screen coordinates
        translatedSelection = new Rect();  //The selected area, screen coordinates
        aspectRatio = 1;
    }
    
    public Selection(Selection sel) {
        setSelection(sel);
    }
    
    public Rect getSelection() {
        return (new Rect(selection));
    }
    
    public Rect getTranslatedSelection() {
        return (new Rect(translatedSelection));
    }
    
    public void setAspectRatio(double x) {
        aspectRatio = x;
    }
    
    public void setLogicalBounds(Rect rect) {
        logicalBounds = new Rect(rect);
        forceBounds();
        translateSelection();
    }
    
    public void setScreenBounds(Rect rect) {
        screenBounds = new Rect(rect);
        forceBounds();
        translateSelection();
    }
    
    public void setSelection(Rect rect) {
        selection = new Rect(rect);
        forceBounds();
        selection.applyAspectRatio(aspectRatio);
        translateSelection();
    }

    public void setSelection(Selection sel) {
        selection = new Rect(sel.selection);
        logicalBounds = new Rect(sel.logicalBounds);
        screenBounds = new Rect(sel.screenBounds);
        translatedSelection = new Rect(sel.translatedSelection);
        aspectRatio = sel.aspectRatio;
    }
    
    public void repaint(Graphics g) {
        g.setColor(new Color(0xffffff));
        g.drawLine((int)translatedSelection.getXmin(), (int)screenBounds.getYmin(),
                   (int)translatedSelection.getXmin(), (int)screenBounds.getYmax()-1);
        g.drawLine((int)translatedSelection.getXmax(), (int)screenBounds.getYmin(),
                   (int)translatedSelection.getXmax(), (int)screenBounds.getYmax()-1);
        g.drawLine((int)screenBounds.getXmin(), (int)translatedSelection.getYmin(),
                   (int)screenBounds.getXmax()-1, (int)translatedSelection.getYmin());
        g.drawLine((int)screenBounds.getXmin(), (int)translatedSelection.getYmax(),
                   (int)screenBounds.getXmax()-1, (int)translatedSelection.getYmax());
    }

    public void moveUsingScreenCoordinates(int x, int y) {
        translatedSelection.add(new Rect(x, x, y, y));
        
        selection = screenToLogical(translatedSelection);
        selection.applyAspectRatio(aspectRatio);
        forceBoundsMaintainingAspect();
        translateSelection();
    }

    /* Resizes the selection with its centre as pivot point */
    public void resizeUsingScreenCoordinates(int x) {
        double middleX, middleY;
        middleX = selection.getX() + selection.getWidth()/2;
        middleY = selection.getY() + selection.getHeight()/2;
        
        translatedSelection.add(new Rect(-x, x, x, -x));
        selection = screenToLogical(translatedSelection);
        forceBounds();
        selection.applyAspectRatio(aspectRatio);
        selection.setPos(middleX - selection.getWidth()/2,
                         middleY - selection.getHeight()/2);
        translateSelection();
    }
    
    private void translateSelection() {
        translatedSelection.setXmin( ((selection.getXmin()-logicalBounds.getXmin()) / logicalBounds.getWidth()) *
                                     (screenBounds.getWidth()-1) + screenBounds.getXmin()
                                   );
        translatedSelection.setXmax( ((selection.getXmax()-logicalBounds.getXmin()) / logicalBounds.getWidth()) *
                                     (screenBounds.getWidth()-1) + screenBounds.getXmin()
                                   );
        translatedSelection.setYmin( (screenBounds.getHeight()-1) - ((selection.getYmin()-logicalBounds.getYmin()) / logicalBounds.getHeight()) *
                                     (screenBounds.getHeight()-1) + screenBounds.getYmin()
                                   );
        translatedSelection.setYmax( (screenBounds.getHeight()-1) - ((selection.getYmax()-logicalBounds.getYmin()) / logicalBounds.getHeight()) *
                                     (screenBounds.getHeight()-1) + screenBounds.getYmin()
                                   );

    }

    /* Makes sure the selection is within its permited area. May
     * spoil the aspect. */
    private void forceBounds() {
        if (selection.getXmin() < logicalBounds.getXmin())
            selection.setXmin(logicalBounds.getXmin());
        if (selection.getXmax() < logicalBounds.getXmin())
            selection.setXmax(logicalBounds.getXmin());
        if (selection.getXmin() > logicalBounds.getXmax())
            selection.setXmin(logicalBounds.getXmax());
        if (selection.getXmax() > logicalBounds.getXmax())
            selection.setXmax(logicalBounds.getXmax());
        
        if (selection.getYmin() < logicalBounds.getYmin())
            selection.setYmin(logicalBounds.getYmin());
        if (selection.getYmax() < logicalBounds.getYmin())
            selection.setYmax(logicalBounds.getYmin());
        if (selection.getYmin() > logicalBounds.getYmax())
            selection.setYmin(logicalBounds.getYmax());
        if (selection.getYmax() > logicalBounds.getYmax())
            selection.setYmax(logicalBounds.getYmax());
        
        if (selection.getXmax() < selection.getXmin())
            selection.setXmax(selection.getXmin());
        if (selection.getXmin() > selection.getXmax())
            selection.setXmin(selection.getXmax());
        if (selection.getYmax() < selection.getYmin())
            selection.setYmax(selection.getYmin());
        if (selection.getYmin() > selection.getYmax())
            selection.setYmin(selection.getYmax());
    }

    /* Moves the selection so that it is completely inside its permited
     boundaries */
    private void forceBoundsMaintainingAspect() {
        if (selection.getXmin() < logicalBounds.getXmin()) {
            selection.setXmax(selection.getXmax() + logicalBounds.getXmin() - selection.getXmin());
            selection.setXmin(logicalBounds.getXmin());
        }
        if (selection.getXmax() > logicalBounds.getXmax()) {
            selection.setXmin(selection.getXmin() + logicalBounds.getXmax() - selection.getXmax());
            selection.setXmax(logicalBounds.getXmax());
        }
        
        if (selection.getYmin() < logicalBounds.getYmin()) {
            selection.setYmax(selection.getYmax() + logicalBounds.getYmin() - selection.getYmin());
            selection.setYmin(logicalBounds.getYmin());
        }
        if (selection.getYmax() > logicalBounds.getYmax()) {
            selection.setYmin(selection.getYmin() + logicalBounds.getYmax() - selection.getYmax());
            selection.setYmax(logicalBounds.getYmax());
        }
    }
    
    private Rect screenToLogical(Rect screen) {
        Rect logical = new Rect();
        
        logical.setXmin( ((screen.getXmin()-screenBounds.getXmin()) / (screenBounds.getWidth()-1)) *
                         (logicalBounds.getWidth()) + logicalBounds.getXmin()
                       );
        logical.setXmax( ((screen.getXmax()-screenBounds.getXmin()) / (screenBounds.getWidth()-1)) *
                         (logicalBounds.getWidth()) + logicalBounds.getXmin()
                       );
        logical.setYmin( (logicalBounds.getHeight()) - ((screen.getYmin()-screenBounds.getYmin()) / (screenBounds.getHeight()-1)) *
                         (logicalBounds.getHeight()) + logicalBounds.getYmin()
                       );
        logical.setYmax( (logicalBounds.getHeight()) - ((screen.getYmax()-screenBounds.getYmin()) / (screenBounds.getHeight()-1)) *
                         (logicalBounds.getHeight()) + logicalBounds.getYmin()
                       );
        
        return (logical);
    }
}
